package com.ibutton.tmex.container;
// iButtonContainer21.java
/*---------------------------------------------------------------------------
 * Copyright (C) 1999 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */
 
// imports
import java.util.StringTokenizer;
import com.ibutton.tmex.adapter.*;
import com.ibutton.tmex.utils.*;

//----------------------------------------------------------------------------
/** iButton container for iButton family type 21 (hex). <p>
 *
 *  History:
 *  <ul>
 *  <li> 0.00 -> 0.01  Initial implementation of download and mission methods.
 *  </ul>
 *
 *  @version    0.01, 25 June 1999
 *  @author     DS,GH
 */
public class iButtonContainer21 extends iButtonContainerGeneral
                                implements iButtonContainer
{
    //-------------------------------------------------------------------------
    //--------Static Final Variables
    //-------------------------------------------------------------------------
    //--------Memory Function Commands
    /** DS1921 Write Scratchpad Command */
    public static final char WRITE_SCRATCHPAD_COMMAND = 0x0F;
    /** DS1921 Read Scratchpad Command*/
    public static final char READ_SCRATCHPAD_COMMAND = 0xAA;
    /** DS1921 Copy Scratchpad Command*/
    public static final char COPY_SCRATCHPAD_COMMAND = 0x55;
    /** DS1921 Read Memory Command*/
    public static final char READ_MEMORY_COMMAND = 0xF0;
    /** DS1921 Read Memory With CRC Command*/
    public static final char READ_MEMORY_WITH_CRC_COMMAND = 0xA5;
    /** DS1921 Clear Memory Command*/
    public static final char CLEAR_MEMORY_COMMAND = 0x3C;
    /** DS1921 Convert Temperature Command*/
    public static final char CONVERT_TEMPERATURE_COMMAND = 0x44;
    
    //--------
    //-------- data variables
    //--------
    /** last read register page */
    char [] regPage = new char [0];
    /** time converter*/
    TimeConverter tc = new TimeConverter(new char[0]);
       
       
    //-------------------------------------------------------------------------
    //-------- Constructor
    //-------------------------------------------------------------------------

    //-------------------------------------------------------------------------
    /** Constructs a General purpose iButton container class.  This 
     *  constructor does not contain a source adapter.  The method
     *  setSourceAdapter must be called right after this method,
     *  before this container can be used.
     */
    public iButtonContainer21()
    {
       super();
    }

    //-------------------------------------------------------------------------
    //-------- Methods
    //-------------------------------------------------------------------------
 
    //--------
    //-------- Information methods
    //--------

    //-------------------------------------------------------------------------
    /** Retrieve the Dallas Semiconductor part number of the iButton 
     *  as a string.  For example 'DS1992'.
     *
     *  @return string represetation of the iButton name.
     */
    public String getiButtonPartName()
    {
       return new String("DS1921");
    }
 
    //-------------------------------------------------------------------------
    /** Retrieve the alternate Dallas Semiconductor part numbers or names.  
     *  A 'family' of MicroLAN devices may have more than one part number
     *  depending on packaging.  There can also be nicknames such as
     *  'Crypto iButton'.
     *
     *  @return string represetation of the alternate names.
     */
    public String getAlternateNames()
    {
       return new String("Thermochron");
    }
 
    //-------------------------------------------------------------------------
    /** Retrieve a short description of the function of the iButton type.
     *
     *  @return string represetation of the function description.
     */
    public String getDescription()
    {
       return new String("Temperature Recorder iButton with digital " + 
                         "thermometer from -40C to 85C in typically " + 
                         "0.2 seconds.  +/- 1C Accuracy between -20C " + 
                         "and 70C.  Also includes a real time clock " +
                         "to wake up and log temperature readings or " + 
                         "collect a long term histogram.");
    } 
   
    //--------
    //-------- Memory Function Command Access Methods
    //--------
    
    //-------------------------------------------------------------------------
    // ReadScratchpad
    /** Read the scratchpad page of memory from the DS1921 
     *  This method reads and returns the entire scratchpad after the byte 
     *  offset regardless of the actual ending offset, in order to do CRC 
     *  verification
     *
     *  @return char array representation of TA1, TA2, e/s byte
     *   and the portion of the 32 byte page after the byte offset.
     *
     *  @exception AdapterNotDetectedException
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
    public char[] readScratchpad()
        throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
    {
        int i; //loop index
        int myCRC; //crc value of returned data
        char[] sendBlock  = new char[38]; //block to send command/read data
        sendBlock[0] = READ_SCRATCHPAD_COMMAND;
        //prepare sendBlock to recieve 37 bytes
        for (i = 1; i < sendBlock.length; i++)
            sendBlock[i] = 0xFF;
            
            
        if (adapter.select(this))
        {
            sendBlock = adapter.dataBlock(sendBlock, 0, sendBlock.length);
        }
        else
        {
            //device must not have been present
            throw new MicroLanIOException("This 1921 not found on Microlan");
        }
        //find offset into scratchpad and get actual return block
        int offset = 0x001F & (int) sendBlock[1]; //lowest 5 bits of address
        char[] crcBlock = new char[sendBlock.length - offset];
        for (i = 0; i<crcBlock.length;  i++)
           crcBlock[i] = sendBlock[i];
               
        //confirm CRC correctness
        CRC16 crc = new CRC16();
        crc.seedCRC16(0);
        myCRC = crc.doCRC16(crcBlock);
        if (myCRC  == 0xB001)
        { 
            //return read page
            char[] returnBlock = new char[crcBlock.length - 3];
            for (i = 0; i<returnBlock.length; i++)
                returnBlock[i] = crcBlock[i+1];
            return returnBlock;
        }
        else 
        {
            //crc check failed
            throw new MicroLanIOException("CRC incorrect");
        }
    }//end readScratchpad()
    
    //-------------------------------------------------------------------------
    // writeScratchpad
    /** Write to the scratchpad page of memory from the DS1921 with CRC verification
     *  the CRC is verified only if the data ends at an offset of 11111b, i.e. 
     *  fills to the end of the page.
     *
     *  @param addr address to write to after verification whern addr[0]= TA2 
     *              and addr[1] = TA1
     *  @param data data to write; only the lower byte of each char is used
     *              and the length of data should not overflow the
     *              page size
     *
     *  @exception AdapterNotDetectedException
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
    public void writeScratchpad(char[] addr, char[] data)
        throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
    {
        int i,j;
        boolean CRCflag = false;
        int byteOffset = (int) addr[1] & 0x001F;
        char [] sendBlock;
        char [] crcBlock;
        
        if ( data.length + byteOffset == 32)
        {
            CRCflag = true;
        }
        else if ( data.length + byteOffset > 32)
        {
            throw new MicroLanIOException("Data would overflow scratchpad;" +
                                          " unable to write to scratchpad.");
        }
        
        //prepare sendBlock
        if (CRCflag)
            sendBlock = new char[3 + data.length + 2];
        else
            sendBlock = new char[3 + data.length];
            
        sendBlock[0] = WRITE_SCRATCHPAD_COMMAND;   //add command
        sendBlock[1] = addr[1];                    //add address TA1, TA2
        sendBlock[2] = addr[0];
        for (i = 0; i<data.length; i++)            //add data bytes
            sendBlock[i+3] = data[i];
        if (CRCflag)                               //if appropriate, add reads for CRC
            for ( j = 0; j<2; j++)
                sendBlock[i+j+3] = 0xFF;
        
        //send data on MicroLan
        if (adapter.select(this))
            sendBlock = adapter.dataBlock(sendBlock, 0, sendBlock.length);
        else
            throw new MicroLanIOException("DS1921 " + getIDString() + " not found on Microlan");
        
        if(CRCflag)
        {
            //confirm CRC correctness
            CRC16 crc = new CRC16();
            crc.seedCRC16(0);
            int myCRC = crc.doCRC16(sendBlock);
              
            if (myCRC != 0xB001)
            { 
                //crc check failed
                throw new MicroLanIOException("CRC incorrect; scratch pad write incorrect");
            }
        }//end if CRCflag
    }
    
    //-------------------------------------------------------------------------
    // copyScratchpad
    /** Copy the scratchpad page to memory.
     *  This method waits 2 microseconds for each byte written to confirm 
     *  that writing has taken place.
     *
     *  @param addr address registers (TA1,TA2,E/S) authorization pattern
     *              It must match exactly the data contained in the three
     *              address registers of the DS1921
     *              addr[0]= TA1
     *              addr[1]= TA2
     *              addr[2]= E/S
     *              Note that a longer array may be used (such as the return 
     *              value from a ReadScratchpad() call) as long as the first
     *              three bytes are correct.
     *
     *  @exception AdapterNotDetectedException
     *  @exception IllegalArgumentException
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
    public void copyScratchpad(char[] addrRegisters)
        throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException,
               IllegalArgumentException
    {
        if (addrRegisters.length < 3)
            throw new IllegalArgumentException("Insufficient register authorization data." +
                                          " TA1, TA2, and E/S required");
        int numBytes = 32 - ( (int) addrRegisters[2] & 0x001F );
        char [] sendBlock = new char[4];
        char returnByte = '\n';
        sendBlock[0] = COPY_SCRATCHPAD_COMMAND;
        sendBlock[1] = addrRegisters[0];
        sendBlock[2] = addrRegisters[1];
        sendBlock[3] = addrRegisters[2];
        
        if (adapter.select(this))
        {
            try
            {
                sendBlock = adapter.dataBlock(sendBlock, 0, sendBlock.length);
                Thread.sleep(0,2000*numBytes); //sleep 2 microsec for every byte to write
                returnByte = getByte();
            }
            catch (InterruptedException ie)
            { }
            catch (IllegalArgumentException e)
            {
                throw new MicroLanIOException("Unable to copy scratchpad");
            }
            
            if (returnByte == 0xFF)
            {
                throw new MicroLanIOException("Authorization failed: unable to copy scratchpad");
            }
            else if ((returnByte == 0x55) || (returnByte == 0xAA))
            { 
                //successful copy
            }
            else 
            {
                throw new MicroLanIOException("Invalid return value on MicroLan");
            }
        }//end if (adapter.select(this))
        else
        {
            //device must not have been present
            throw new MicroLanIOException("This 1921 not found on Microlan");
        }
    }//end copyScratchpad
    
    //-------------------------------------------------------------------------
    //readMemory
    /** reads from memory without CRC verification
     *   @param addr address to start reading from; addr[0]= TA2, addr[1] = TA1
     *   @param numBytes number of bytes to read from this DS1921
     *   @return char[] with the read information in it; each char contains 
     *           a single byte. The address is not returned in the char[]
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
    public char[] readMemory(char[] addr, int numBytes)
        throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
    {
        if (addr.length < 2)
            throw new MicroLanIOException("Address for DS1921 read insufficient");
        
        int i;
        char [] sendBlock = new char[3 + numBytes];
        sendBlock[0] = READ_MEMORY_COMMAND;
        sendBlock[1] = addr[1];
        sendBlock[2] = addr[0];
        for (i=3; i<sendBlock.length; i++)
            sendBlock[i] = 0xFF;
        if (adapter.select(this))
        {
            try
            {
                sendBlock = adapter.dataBlock(sendBlock,0,sendBlock.length);
            }
            catch (Exception e)
            { 
                e.printStackTrace();
                throw new MicroLanIOException("Unable to read from memory");
            }
        }
        else
        {
            //device must not have been present
            throw new MicroLanIOException("this 1921 not found on Microlan");
        }
        char [] returnBlock = new char[numBytes];
        for (i=0; i<returnBlock.length; i++)
            returnBlock[i] = sendBlock[i+3];
        return returnBlock;
    }//end readMemory

    //-------------------------------------------------------------------------
    //  readMemoryWithCRC
    /** Read from memory of the DS1921 with CRC verification
     *  if the initial address is not at the start of a page, the 
     *  page will be read only from that starting offset. This means
     *  that the second page will not start 32 characters into the
     *  return array, but 32-(number of bytes skipped by the offset).
     *  A MicroLanIOException will be thrown if CRC's are incorrect.
     *  @param addr starting address to read from addr[0]= TA2, addr[1] = TA1
     *  @param pagesToRead number of pages to read including the first page
     *  @return char array representation of consecutive bytes read
     *               the address and crc's are not returned as part of the char[]
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
    public char[] readMemoryWithCRC(char[] addr, int pagesToRead)
        throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
    {
        char[] crcBlock;
        char[] sendBlock;
        char[] returnBlock;
        int i,j;
        int myCrcResult;
        int firstPageSize = 32 - ((int) addr[1] & 0x001F);
        CRC16 crc = new CRC16();
        
        if (addr.length > 2)
            throw new MicroLanIOException("Insufficient address for read");
        
        if (pagesToRead <= 0)
            return new char[0];
            
        if (adapter.select(this))
        {
            //need 3 bytes for command and address
            //     firstPageSize +2 bytes for first page
            //     34 bytes for each additional page
            sendBlock = new char[ 5 + firstPageSize + (34*(pagesToRead-1))];
            returnBlock = new char[ firstPageSize + (32 *(pagesToRead-1))];
            sendBlock[0] = READ_MEMORY_WITH_CRC_COMMAND;
            sendBlock[1] = addr[1];
            sendBlock[2] = addr[0];
            for (i=3; i < sendBlock.length; i++)
            {
                sendBlock[i] = 0xFF;
            }
            //send/read data
            sendBlock = adapter.dataBlock(sendBlock, 0, sendBlock.length);
            
            //check CRCs
            //    check first page
            crcBlock = new char[5 + firstPageSize];
            for( i= 0; i < crcBlock.length; i++)
                crcBlock[i] = sendBlock[i];
            crc.seedCRC16(0);
            myCrcResult = crc.doCRC16(crcBlock);
            if (myCrcResult  != 0xB001)
                throw new MicroLanIOException("CRC failed!");
            else 
            {
                for(i = 0; i< (crcBlock.length - 5); i++)
                    returnBlock[i] = crcBlock[i+3];
            }
            
            //    check subsequent pages
            for (i=1; i< pagesToRead; i++)
            {
                crcBlock = new char[34];
                for (j=0; j< crcBlock.length; j++)
                {   
                    crcBlock[j] = sendBlock[5+firstPageSize + (34 * (i-1)) + j];
                }
                crc.seedCRC16(0);
                myCrcResult = crc.doCRC16(crcBlock);
                if (myCrcResult  != 0xB001)
                    throw new MicroLanIOException("CRC failed!");
                else 
                {
                   for(j=0;(j<32) && ((firstPageSize + (32* (i-1))+ j) < returnBlock.length) ;j++)
                   {
                       returnBlock[firstPageSize + (32* (i-1))+ j] = 
                           crcBlock[j];
                   }
                }
             }//end for pagesToRead
        }//end if adapter.select(this)
        else
        {
            //device must not have been present
            throw new MicroLanIOException("this 1921 not found on Microlan");
        }
        return returnBlock;
    }//end readMemoryWithCRC()
    
    //-------------------------------------------------------------------------
    // clearMemory
    /** Clears the memory of the DS1921, usually in preparation to set
     *  up another mission.
     *  This operation takes approximately 500 microseconds to complete
     *  but reset-presence pulses, ROM commands, access to the user RAM,
     *  and reading the clock or status register are possible while
     *  the memory clear takes place. It DOES NOT clear the TLF, THF, and TAF
     *  bits in the status register which must be set to 0 explicitly
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
    public void clearMemory()
        throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
    {
        int i;
        
        //set MCLRE bit to 1 in the control register
        char[] sendBlock = new char[4];
        sendBlock[0] = WRITE_SCRATCHPAD_COMMAND;
        sendBlock[1] = 0x0E;  //TA1
        sendBlock[2] = 0x02;  //TA2
        sendBlock[3] = 0x40;  //Data which sets the MCLRE bit
        
        char[] recieveBlock = new char[5];
        recieveBlock[0] = READ_SCRATCHPAD_COMMAND;
        for (i = 1; i < recieveBlock.length; i++)
            recieveBlock[i] = 0xFF; //read data 
        
        if (adapter.select(this))
            sendBlock = adapter.dataBlock(sendBlock, 0, sendBlock.length);
        else
            throw new MicroLanIOException("Device not on MicroLan");
        if (adapter.select(this)) // reset/presence/match ROM
                recieveBlock = adapter.dataBlock(recieveBlock, 0, recieveBlock.length);
        else
            throw new MicroLanIOException("Device not on MicroLan");
            
        if ((recieveBlock[1] != 0x0E) ||
            (recieveBlock[2] != 0x02) ||
            ((recieveBlock[3] & 0x1F) != 0x0E) || //mask out flags
            (recieveBlock[4] != 0x40))
            throw new MicroLanIOException("Data Transfer Unconfirmed");
            
        char[] confirmBlock = new char[5];
        confirmBlock[0] = COPY_SCRATCHPAD_COMMAND;
        confirmBlock[1] = recieveBlock[1];
        confirmBlock[2] = recieveBlock[2];
        confirmBlock[3] = recieveBlock[3];
        
        if (adapter.select(this))
            confirmBlock = adapter.dataBlock(confirmBlock, 0, confirmBlock.length);
        else 
            throw new MicroLanIOException("Device not on MicroLan");
            
        //send the CLEAR_MEMORY_COMMAND
        if (adapter.select(this))
            adapter.putByte(CLEAR_MEMORY_COMMAND);
        else 
            throw new MicroLanIOException("Device not on MicroLan");
        
        //read MEMCLR bit to confirm that memory has been cleared
        confirmBlock = new char[4];
        confirmBlock[0] = READ_MEMORY_COMMAND;
        confirmBlock[1] = 0x14; //address of status byte
        confirmBlock[2] = 0x02;
        confirmBlock[3] = 0xFF; //position for read data
        
        if(adapter.select(this))
            confirmBlock = adapter.dataBlock(confirmBlock, 0, confirmBlock.length);
        else
            throw new MicroLanIOException("Device not on MicroLan");
            
        if (((int)confirmBlock[3] & 0x40) != 0x40) // MEMCLR not set
            throw new MicroLanIOException("Memory not cleared");
   }// end clearMemory
    
    //-------------------------------------------------------------------------
    // forceConvertTemperature
    /** If a mission is not in progress, does a temperature conversion, the
     *  results of which are returned and also may be found in the 0x211 
     *  address within the register page.
     *  @return the celsius value of the temperature measured
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
    public float forceConvertTemperature()
        throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
    {
        //confirm that there is currently no mission
        char[] getStatusBlock = new char[4];
        getStatusBlock[0] = READ_MEMORY_COMMAND;
        getStatusBlock[1] = 0x14;
        getStatusBlock[2] = 0x02;
        getStatusBlock[3] = 0xFF;
        
        if (adapter.select(this))
            getStatusBlock = adapter.dataBlock(getStatusBlock, 0, getStatusBlock.length);
        else
            throw new MicroLanIOException("Device not prestent");
        
        if ( (getStatusBlock[3] & 0x20) == 0x20) //MIP == 1
            throw new MicroLanIOException("Device in Mission; Function Unavailable");
            
        //send temperature command
        if (adapter.select(this))
            adapter.putByte(CONVERT_TEMPERATURE_COMMAND);
        else 
            throw new MicroLanIOException("Device not prestent");
        //wait for the reading to finish
        try
        {
            Thread.sleep(750);
        }
        catch (Exception e)
        { }
        
        //read the register page
        char[] getReadingBlock = new char[4];
        getReadingBlock[0] = READ_MEMORY_COMMAND;
        getReadingBlock[1] = 0x11; 
        getReadingBlock[2] = 0x02;
        getReadingBlock[3] = 0xFF;
        
        if (adapter.select(this))
            getReadingBlock = adapter.dataBlock(getReadingBlock, 0, getReadingBlock.length);
        else
            throw new MicroLanIOException("Device not present");
        
        //return the read temperature
        float temperature = (float)getReadingBlock[3]/2 - 40; //convert to celsius
        return temperature;
    }
    
    //-------
    //------- Register Page Information Access Methods
    //-------
    
    //-------------------------------------------------------------------------
    // getRegisterPage
    /** retrieves and returns as a char[] the registerpage of this 1921
     *  @return the register page as a char[], where each char contains
     *   a single byte.
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public char[] getRegisterPage()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        char[] sendBlock = new char[37]; // 1 command+ 2 address + 32 data + 2 CRC
        char[] returnBlock = new char[32]; //32 data bytes
        CRC16 crc = new CRC16();
        int myCrcValue;
        int i;
        
        sendBlock[0] = READ_MEMORY_WITH_CRC_COMMAND;
        sendBlock[1] = 0x00; //register page begins at 0200h
        sendBlock[2] = 0x02;
        for (i=3; i< sendBlock.length; i++)
            sendBlock[i] = 0xFF;
        
        if (adapter.select(this))
            sendBlock = adapter.dataBlock(sendBlock, 0, sendBlock.length);
        else
            throw new MicroLanIOException("DS1921 " + this.getIDString() + " not found on 1-Wire net");
            
        crc.seedCRC16(0);
        myCrcValue = crc.doCRC16(sendBlock);
        
        if (myCrcValue != 0xB001)
            throw new MicroLanIOException("CRC incorrect");
            
        for (i = 0; i < returnBlock.length; i++)
            returnBlock[i] = sendBlock[i+3];
            
        regPage = returnBlock; 
        return returnBlock;
     }//end getRegisterPage
        
    
     
    //-------------------------------------------------------------------------
    // getLowTemperatureThreshold
    /** retrieves and returns the low temperature threshold from this 1921. 
     *  @return the temperature threshold in degrees celsius
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public float getLowTemperatureThreshold()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
         getRegisterPage();
         return getLowTemperatureThreshold(regPage);
     }//end getLowTemperatureThreshold
     
     //-------------------------------------------------------------------------
    // getLowTemperatureThreshold
    /** retrieves and returns the low temperature threshold from this 1921. 
     *  @param registerPage char[] containing a copy of the registerPage from
     *   which the temperature threshold is determined.
     *  @see #getRegisterPage()
     *  @return the temperature threshold in degrees celsius
     */
     public static float getLowTemperatureThreshold(char[] registerPage)
     {
        return (((float) registerPage[11])/2) -40;
     }//end getLowTemperatureThreshold
     
     
     //-------------------------------------------------------------------------
    // getHighTemperatureThreshold
    /** retrieves and returns the High temperature threshold from this 1921. 
     *  @return the temperature threshold in degrees celsius
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public float getHighTemperatureThreshold()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
         getRegisterPage();
         return getHighTemperatureThreshold(regPage);
     }//end getHighTemperatureThreshold
     
    //-------------------------------------------------------------------------
    // getHighTemperatureThreshold
    /** retrieves and returns the High temperature threshold from this 1921. 
    *  @param registerPage char[] containing a copy of the registerPage from
     *   which the temperature threshold is determined.
     *  @see #getRegisterPage()
     *  @return the temperature threshold in degrees celsius
     */
     public static float getHighTemperatureThreshold(char[] registerPage)
     {
        return (((float) registerPage[12])/2) -40;
     }//end getHighTemperatureThreshold
     
    //-------------------------------------------------------------------------
    // getSampleRate
    /** retrieves and returns the number of minutes between samples from this 1921. 
     *  @return the number of minutes between samples
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public int getSampleRate()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        getRegisterPage();
        return getSampleRate(regPage);
     }//end getSampleRate
     
    //-------------------------------------------------------------------------
    // getSampleRate
    /** retrieves and returns the number of minutes between samples from this 1921. 
    *  @param registerPage char[] containing a copy of the registerPage from
     *   which the sample rate is determined.
     *  @see #getRegisterPage()
     *  @return the number of minutes between samples
     */
     public static int getSampleRate(char[] registerPage)
     {
        return (int) registerPage[13];
     }//end getSampleRate
    
    //-------------------------------------------------------------------------
    // getMissionSamplesCounter
    /** retrieves and returns the number of mission samples from this 1921. 
     *  @return the number of mission samples
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public int getMissionSamplesCounter()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        getRegisterPage();
        return getMissionSamplesCounter(regPage);
     }//end getMissionSamplesCounter
     
     //-------------------------------------------------------------------------
    // getMissionSamplesCounter
    /** retrieves and returns the number of mission samples from this 1921. 
    *  @param registerPage char[] containing a copy of the registerPage from
     *   which the mission sample value is determined.
     *  @see #getRegisterPage()
     *  @return the number of mission samples
     */
     public static int getMissionSamplesCounter(char[] registerPage)
     {
        int samples = 0;
        samples = samples + registerPage[28];
        samples = (samples << 8) | registerPage[27];
        samples = (samples << 8) | registerPage[26];
        return samples;
     }//end getMissionSamplesCounter
     
    //-------------------------------------------------------------------------
    // getDeviceSamplesCounter
    /** retrieves and returns the number of device samples from this 1921. 
     *  @return the number of device samples
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public int getDeviceSamplesCounter()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        getRegisterPage();
        return getDeviceSamplesCounter(regPage);
     }//end getDeviceSamplesCounter
     
     
     //-------------------------------------------------------------------------
    // getDeviceSamplesCounter
    /** retrieves and returns the number of device samples from this 1921. 
     *  @param registerPage char[] containing a copy of the registerPage from
     *   which the device sample value is determined.
     *  @see #getRegisterPage()
     *  @return the number of device samples
     */
     public static int getDeviceSamplesCounter(char[] registerPage)
     {
        int samples = 0;
        samples = samples + registerPage[31];
        samples = (samples << 8) + registerPage[30];
        samples = (samples << 8) + registerPage[29];
        return samples;
     }//end getDeviceSamplesCounter
     
    //-------------------------------------------------------------------------
    // getMissionStartDelay
    /** retrieves and returns the number of minutes delay from mission start to 
     *  first temperature reading from this 1921. 
     *  @return the number of minutes delay
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public int getMissionStartDelay()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        getRegisterPage();
        return getMissionStartDelay(regPage);
     }//end getMissionStartDelay()
     
    //-------------------------------------------------------------------------
    // getMissionStartDelay
    /** retrieves and returns the number of minutes delay from mission start to 
     *  first temperature reading from this 1921. 
     *  @param registerPage char[] containing a copy of the registerPage from
     *   which the mission start delay is determined.
     *  @see #getRegisterPage()
     *  @return the number of minutes delay
     */
     public static int getMissionStartDelay(char[] registerPage)
     {
        int min = 0;
        min = min + registerPage[19];
        min = (min << 8) + registerPage[18];
        return min;
     }//end getMissionStartDelay()
     
    //-------------------------------------------------------------------------
    // createMissionStartDelay
    /** create mission start delay char[] of length 2 for sending to the 
     *  DS1921 in a new mission
     *  @param msd number of minutes before the mission  starts. Must be in the
     *  range from 0 to 65535
     *  @see #mission(char[] missionPage, boolean newClock)
     *  @return char[] containing the start delay in DS1921 format. 
     *  element [0] is the low byte.
     *  
     *  @exception IllegalArgumentException
     */
     public char[] createMissionStartDelay(int msd)
         throws IllegalArgumentException
     {
        if ( (msd <0) || (msd > 65535))
            throw new IllegalArgumentException("Mission Start Delay Out Of Bounds. Range is 0-65535.");
        char[] msdChar = new char[2];
        msdChar[0]=(char)(msd & 0x00FF);
        msdChar[1]=(char)((msd & 0xFF00) >>> 8);
        return msdChar;
     }//end createMissionStartDelay() 
     
    //-------------------------------------------------------------------------
    // getLastTempReading
    /** retrieves and returns the celsius temperature in the temperature read out 
     *  of this 1921. NOTE: does NOT do a current temperature reading, but retrieves
     *  the result of the last one
     *  @return the temperature in the temperature read out
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public float getLastTempReading()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        getRegisterPage();
        return getLastTempReading(regPage);
     }
     
    //-------------------------------------------------------------------------
    // getLastTempReading
    /** retrieves and returns the celsius temperature in the temperature read out 
     *  of this 1921. NOTE: does NOT do a current temperature reading, but retrieves
     *  the result of the last one
     *  @param registerPage char[] containing a copy of the registerPage from
     *   which the temperatuer reading is determined.
     *  @see #getRegisterPage()
     *  @return the temperature in the temperature read out 
     *
     */
     public static float getLastTempReading(char[] registerPage)
     {
        return ((int) registerPage[17]/2)-40;
     }
     
    
     
    //-------------------------------------------------------------------------
    // getControlRegister
    /** retrieves and returns the control register of this 1921. 
     *  @return the control register in a char. Only the lower 8 bits are used.
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public char getControlRegister()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        getRegisterPage();
        return getControlRegister(regPage);
     }
     
    //-------------------------------------------------------------------------
    // getControlRegister
    /** retrieves and returns the control register of this 1921. 
     *  @param registerPage char[] containing a copy of the registerPage from
     *   which the control register is determined.
     *  @see #getRegisterPage()
     *  @return the control register in a char. Only the lower 8 bits are used.
     */
     public static char getControlRegister(char[] registerPage)
     {
        return registerPage[14];
     }
     
    //-------------------------------------------------------------------------
    // getControlByteForNewMission
    /** Creates a control byte for a new mission. NEOSC is set to 0, MCLRE = 0,
     *  NEM = 0, and the RO, TLS, THS, and TAS bits are set to the values given.
     *  @param RO true=enable rollover, false=disable rollover
     *  @param TLS true=respond to conditional search if a low temp alarm has 
     *  occurred
     *  @param THS true=respond to conditional search if a high temp alarm has
     *  occurred
     *  @param TAS true=respond to conditional search if a real time clock alarm
     *  has occurred
     */
    public char createNewControlRegister(boolean RO, boolean TLS, boolean THS, boolean TAS)
    {
       char control = 0x00;
       if (RO)
          control = (char)(control | 0x08);
       if (TLS)
          control = (char)(control | 0x04);
       if (THS)
          control = (char)(control | 0x02);
       if (TAS)
          control = (char)(control | 0x01);
       return control;
    }//end createNewControlRegister

    //-------------------------------------------------------------------------
    // getControlBitTAS
    /** retrieves and returns the Time Alarm Search of this 1921.
     *  TAS true will respond to conditional search if a timer alarm has occurred. 
     *  False will not respond to conditional search if a timer alarm has occurred.
     *  @return Timer Alarm Search value
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public boolean getControlBitTAS()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        getRegisterPage();
        return getControlBitTAS(regPage);
     }
     
    //-------------------------------------------------------------------------
    // getControlBitTAS
    /** retrieves and returns the Time Alarm Search of the given register page.
     *  TAS true will respond to conditional search if a timer alarm has occurred. 
     *  False will not respond to conditional search if a timer alarm has occurred.
     *  @param registerPage char[] containing a copy of the registerPage from
     *   which the control bit is determined.
     *  @see #getRegisterPage()
     *  @return Timer Alarm Search value
     */
     public static boolean getControlBitTAS(char[] registerPage)
     {
        return ( (registerPage[14] & 0x01) == 0x01);
     }
     
    //-------------------------------------------------------------------------
    // getControlBitTHS
    /** retrieves and returns the Temperature High Alarm Search of this 1921.
     *  THS true will respond to conditional search if a temp high alarm has
     *  occurred. False will not respond to conditional search if a temp high 
     *  alarm has occurred.
     *  @return Temperature High Alarm Search value
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public boolean getControlBitTHS()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        getRegisterPage();
        return getControlBitTHS(regPage);
     }
     
    //-------------------------------------------------------------------------
    // getControlBitTHS
    /** retrieves and returns the Temperature High Alarm Search of the given
     *  register page.
     *  THS true will respond to conditional search if a temp high alarm has
     *  occurred. False will not respond to conditional search if a temp high 
     *  alarm has occurred.
     *  @param registerPage char[] containing a copy of the registerPage from
     *   which the control bit is determined.
     *  @see #getRegisterPage()
     *  @return Temperature High Alarm Search value
     */
     public static boolean getControlBitTHS(char[] registerPage)
     {
        return ( (registerPage[14] & 0x02) == 0x02);
     }
     
    //-------------------------------------------------------------------------
    // getControlBitTLS
    /** retrieves and returns the Temperature Low Alarm Search of this 1921.
     *  TLS true will respond to conditional search if a temp low alarm has 
     *  occured. False will not respond to conditional search if a temp low 
     *  alarm has occured. 
     *  @return Temperature Low Alarm Search value
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public boolean getControlBitTLS()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        getRegisterPage();
        return getControlBitTLS(regPage);
     }
     
    //-------------------------------------------------------------------------
    // getControlBitTLS
    /** retrieves and returns the Temperature Low Alarm Search from the given
     *  register page.
     *  TLS true will respond to conditional search if a temp low alarm has 
     *  occured. False will not respond to conditional search if a temp low 
     *  alarm has occured. 
     *  @param registerPage char[] containing a copy of the registerPage from
     *   which the control bit is determined.
     *  @see #getRegisterPage()
     *  @return Temperature Low Alarm Search value
     */
     public static boolean getControlBitTLS(char[] registerPage)
     {
        return ( (registerPage[14] & 0x04) == 0x04);
     }
     
    //-------------------------------------------------------------------------
    // getControlBitRO
    /** retrieves and returns the Roll-Over Enable/Disable setting of this 1921.
     *  RO true indicates roll-over is enabled; when the temperature logging 
     *  memory is full, new readings will overwrite the old ones. False indicates 
     *  that roll-over is not enabled; when the 96 pages of temperature logging 
     *  memory, the initial records will not be over-written by later readings.
     *  @return Roll-Over Enable/Disable value
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public boolean getControlBitRO()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        getRegisterPage();
        return getControlBitRO(regPage);
     }
     
    //-------------------------------------------------------------------------
    // getControlBitRO
    /** retrieves and returns the Roll-Over Enable/Disable setting from the given
     *  register page.
     *  RO true indicates roll-over is enabled; when the temperature logging 
     *  memory is full, new readings will overwrite the old ones. False indicates 
     *  that roll-over is not enabled; when the 96 pages of temperature logging 
     *  memory, the initial records will not be over-written by later readings.
     *  @param registerPage char[] containing a copy of the registerPage from
     *   which the control bit is determined.
     *  @see #getRegisterPage()
     *  @return Roll-Over Enable/Disable value
     */
     public static boolean getControlBitRO(char[] registerPage)
     {
        return ( (registerPage[14] & 0x08) == 0x08);
     }
     
    //-------------------------------------------------------------------------
    // getControlBitEM
    /** retrieves and returns the Enable Mission setting of this 1921.
     *  EM false indicates that a new mission is enabled and will start
     *  when a non-zero sample rate is written to the device. EM true indicates 
     *  that the device is not enabled for a mission.
     *  @return Enable Mission value
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public boolean getControlBitEM()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        getRegisterPage();
        return getControlBitEM(regPage);
     }
     
    //-------------------------------------------------------------------------
    // getControlBitEM
    /** retrieves and returns the Enable Mission setting from the given register
     *  page.
     *  EM false indicates that a new mission is enabled and will start
     *  when a non-zero sample rate is written to the device. EM true indicates 
     *  that the device is not enabled for a mission.
     *  @param registerPage char[] containing a copy of the registerPage from
     *   which the control bit is determined.
     *  @see #getRegisterPage()
     *  @return EM value
     */
     public static boolean getControlBitEM(char[] registerPage)
     {
        return ( (registerPage[14] & 0x10) == 0x10);
     }
     
    //-------------------------------------------------------------------------
    // getControlBitMCLRE
    /** retrieves and returns the Memory Clear Enable setting of this 1921.
     *  MCLRE true indicates that the device will clear the mission related 
     *  memory if the next command is Clear Memory. False indicates that clearing
     *  the memory is not enabled.
     *  @return Memory Clear Enable value
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public boolean getControlBitMCLRE()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        getRegisterPage();
        return getControlBitMCLRE(regPage);
     }
     
    //-------------------------------------------------------------------------
    // getControlBitMCLRE
    /** retrieves and returns the Memory Clear Enable setting from the given
     *  register page.
     *  MCLRE true indicates that the device will clear the mission related 
     *  memory if the next command is Clear Memory. False indicates that clearing
     *  the memory is not enabled.
     *  @param registerPage char[] containing a copy of the registerPage from
     *   which the control bit is determined.
     *  @see #getRegisterPage()
     *  @return Memory Clear Enable value
     */
     public boolean getControlBitMCLRE(char [] registerPage)
     {
        return ( (registerPage[14] & 0x40) == 0x40);
     }
     
    //-------------------------------------------------------------------------
    // getControlBitEOSC
    /** retrieves and returns the Enable Oscillator of this 1921.
     *  EOSC false indicates the the oscillator is enabled and therefore the real
     *  time clock is running. True indicates that the oscillator is disabled, and
     *  therefore the device is in low-power data retention mode.
     *  @return Enable Oscillator value
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public boolean getControlBitEOSC()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
       getRegisterPage();
        return getControlBitEOSC(regPage);
     }
     
    //-------------------------------------------------------------------------
    // getControlBitEOSC
    /** retrieves and returns the Enable Oscillator of this 1921.
     *  EOSC false indicates the the oscillator is enabled and therefore the real
     *  time clock is running. True indicates that the oscillator is disabled, and
     *  therefore the device is in low-power data retention mode.
     *  @param registerPage char[] containing a copy of the registerPage from
     *   which the control bit is determined.
     *  @see #getRegisterPage()
     *  @return Enable Oscillator value
     */
     public boolean getControlBitEOSC(char [] registerPage)
     {
        return ( (registerPage[14] & 0x80) == 0x80);
     }
     
    //-------------------------------------------------------------------------
    // getStatusRegister
    /** retrieves and returns the status register of this 1921. 
     *  @return the status register in a char. Only the lower 8 bits are used.
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public char getStatusRegister()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        getRegisterPage();
        return getStatusRegister(regPage);
     }
     
     //-------------------------------------------------------------------------
    // getStatusRegister
    /** retrieves and returns the status register of this 1921. 
     *  @param registerPage char[] containing a copy of the registerPage from
     *   which the status register is determined.
     *  @see #getRegisterPage()
     *  @return the status register in a char. Only the lower 8 bits are used.
     */
     public static char getStatusRegister(char [] registerPage)
     {
         return registerPage[20];
     }

    //-------------------------------------------------------------------------
    // getStatusBitTAF
    /** retrieves and returns the Time Alarm Flag of this 1921.
     *  TAF true indicates that a clock alarm has occured. False indicates that 
     *  no clock alarm has yet occured during this mission.
     *  @return Time Alarm Flag value
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public boolean getStatusBitTAF()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        getRegisterPage();
        return getStatusBitTAF(regPage);
     }
     
    //-------------------------------------------------------------------------
    // getStatusBitTAF
    /** retrieves and returns the Time Alarm Flag from the given register page.
     *  TAF true indicates that a clock alarm has occured. False indicates that 
     *  no clock alarm has yet occured during this mission.
     *  @param registerPage char[] containing a copy of the registerPage 
     *  @see #getRegisterPage()
     *  @return Time Alarm Flag value
     */
     public static boolean getStatusBitTAF(char[] registerPage)
     {
        return (registerPage[20] & 0x01) == 0x01;
     }
     
    //-------------------------------------------------------------------------
    // getStatusBitTHF
    /** retrieves and returns the Temperature High Flag of this 1921.
     *  THF true indicates that a temperature high alarm has occured. False 
     *  indicates that no temperature high alarm has yet occured during this 
     *  mission. 
     *  @return Temperature High Flag value
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public boolean getStatusBitTHF()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        getRegisterPage();
        return getStatusBitTHF(regPage);
     }
     
    //-------------------------------------------------------------------------
    // getStatusBitTHF
    /** retrieves and returns the Temperature High Flag from the given register page.
     *  THF true indicates that a temperature high alarm has occured. False 
     *  indicates that no temperature high alarm has yet occured during this mission. 
     *  @param registerPage char[] containing a copy of the registerPage 
     *  @see #getRegisterPage()
     *  @return Temperature High Flag value
     */
     public static boolean getStatusBitTHF(char[] registerPage)
     {
        return (registerPage[20] & 0x02) == 0x02;
     }
     
    //-------------------------------------------------------------------------
    // getStatusBitTLF
    /** retrieves and returns the Temperature Low Flag of this 1921.
     *  TLF true indicates that a temperature low alarm has occured. False
     *  indicates that no temperature low alarm has occured during this 
     *  mission.
     *  @return Temperatuer Low Flag value
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public boolean getStatusBitTLF()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        getRegisterPage();
        return getStatusBitTLF(regPage);
     }
     
    //-------------------------------------------------------------------------
    // getStatusBitTLF
    /** retrieves and returns the Temperature Low Flag of the given register page.
     *  TLF true indicates that a temperature low alarm has occured. False
     *  indicates that no temperature low alarm has occured during this 
     *  mission.
     *  @param registerPage char[] containing a copy of the registerPage 
     *  @see #getRegisterPage()
     *  @return Temperature  Low Flag value
     */
     public static boolean getStatusBitTLF(char[] registerPage)
     {
        return (registerPage[20] & 0x04) == 0x04;
     }
     
    //-------------------------------------------------------------------------
    // getStatusBitSIP
    /** retrieves and returns the Sample In Progress flag of this 1921.
     *  SIP true indicates that a self initiated temperature reading
     *  as part of a mission is taking place. False indicates that no mission
     *  initiated temperture reading is currently in progress.
     *  @return Sample In Progress value
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public boolean getStatusBitSIP()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        getRegisterPage();
        return getStatusBitSIP(regPage);
     }
     
    //-------------------------------------------------------------------------
    // getStatusBitSIP
    /** retrieves and returns the Sample In Progress flag from the given
     *  register page.
     *  SIP true indicates that a self initiated temperature reading
     *  as part of a mission is taking place. False indicates that no mission
     *  initiated temperture reading is currently in progress.
     *  @param registerPage char[] containing a copy of the registerPage from
     *   which the status bit is determined.
     *  @see #getRegisterPage()
     *  @return Sample In Progress value
     */
     public static boolean getStatusBitSIP(char[] registerPage)
     {
        return ((registerPage[20] & 0x10) == 0x10);
     }
     
    //-------------------------------------------------------------------------
    // getStatusBitMIP
    /** retrieves and returns the Misson In Progress flag of this 1921.
     *  MIP true indicates that a mission is in progress. False indicates that
     *  no mission is currently running.
     *  @return Mission In Progress value
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public boolean getStatusBitMIP()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        getRegisterPage();
        return getStatusBitMIP(regPage);
     }
     
    //-------------------------------------------------------------------------
    // getStatusBitMIP
    /** retrieves and returns the Misson In Progress flag from the given register
     *  page.
     *  MIP true indicates that a mission is in progress. False indicates that
     *  no mission is currently running.
     *  @param registerPage char[] containing a copy of the registerPage from
     *   which the status bit is determined.
     *  @see #getRegisterPage()
     *  @return MIP value
     */
     public static boolean getStatusBitMIP(char[] registerPage)
     {
        return ((registerPage[20]& 0x20) == 0x20);
     }
     
    //-------------------------------------------------------------------------
    // getStatusBitMEMCLR
    /** retrieves and returns the Memory Cleared Flag of this 1921.
     *  MEMCLR true indicates that all memory pages 17 and higher along with other
     *  mission related data has been cleared to 0 with the Clear Memory Command. 
     *  If any data has been written to the those locations, the MEMCLR value is 
     *  false.
     *  @return Memory Cleared value
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public boolean getStatusBitMEMCLR()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        getRegisterPage();
        return getStatusBitMEMCLR(regPage);
     }
     
    //-------------------------------------------------------------------------
    // getStatusBitMEMCLR
    /** retrieves and returns the Memory Cleared Flag from the given register
     *  page
     *  MEMCLR true indicates that all memory pages 17 and higher along with other
     *  mission related data has been cleared to 0 with the Clear Memory Command. 
     *  If any data has been written to the those locations, the MEMCLR value is 
     *  false.
     *  @param registerPage char[] containing a copy of the registerPage from
     *   which the status bit is determined.
     *  @see #getRegisterPage()
     *  @return Memory Cleared value
     */
     public static boolean getStatusBitMEMCLR(char[] registerPage)
     {
        return ((registerPage[20] & 0x40) == 0x40);
     }
     
    //-------------------------------------------------------------------------
    // getStatusBit7
    /** retrieves and returns status bit 7 of this 1921.
     *  Status bit 7 false indicates that the temperature core is busy performing 
     *  a conversion. True indicates that no temperature sample is currently
     *  in progress.
     *  @return status bit 7 value
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public boolean getStatusBit7()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        getRegisterPage();
        return getStatusBit7(regPage);
     }
     
    //-------------------------------------------------------------------------
    // getStatusBit7
    /** retrieves and returns status bit 7 of this 1921.
     *  Status bit 7 false indicates that the temperature core is busy performing 
     *  a conversion. True indicates that no temperature sample is currently
     *  in progress.
     *  @param registerPage char[] containing a copy of the registerPage from
     *   which the status bit is determined.
     *  @see #getRegisterPage()
     *  @return status bit 7 value
     */
     public static boolean getStatusBit7(char[] registerPage)
     {
        return((registerPage[20] & 0x80) == 0x80);
     }
         
    //--------
    //-------- Time Format Access Methods
    //--------
    
    //-------------------------------------------------------------------------
    //toRTCTime
    /** creates a Real Time Clock array block from the time string
     *   @param time time string in the format DayOfWeek mm/dd/yyyy hh:mm:ss
     *   @return char[] containing Real Time Clock information in the 
     *   DS1921 format
     *   @exception IllegalArgumentException
     */
    public char[] toRTCTime(String time)
        throws IllegalArgumentException 
    {
        return tc.toRTCTime(time);
    }
    
    //-------------------------------------------------------------------------
    //toRTCATime
    /** creates a Real Time Clock Alarm array block from the time string
     *   @param time time string in the format DayOfWeek hh:mm:ss
     *   @param alarmMask alarm mask which determines the frequency of 
     *   the alarm timer
     *   @return char[4] containing Real Time Clock Alarm information in the 
     *   DS1921 format
     *   @exception IllegalArgumentException
     */   
    public char[] toRTCATime(String time, char alarmMask)
        throws IllegalArgumentException 
    {
        return tc.toRTCATime(time, alarmMask);
    }
     
    
     
     //-------------------------------------------------------------------------
    // getRealTimeClockString
    /** retrieves and returns the current time from this 1921. The return string
     *  is of the form dayofweek mm/dd/yyyy hh:mm:ss
     *              or dayofweek mm/dd/yyyy hh:mm:ss AM/PM 
     *  depending on the 12/24 format of the iButton. For example:
     *  Wednesday 5/5/1999 3:39:34 PM
     *      
     *  @return a String containing the current time
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public String getRealTimeClockString()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
         getRegisterPage();
         return getRealTimeClockString(regPage);
     }
     
    //-------------------------------------------------------------------------
    // getRealTimeClockString
    /** retrieves and returns the current time from this 1921. The return string
     *  is of the form dayofweek mm/dd/yyyy hh:mm:ss
     *              or dayofweek mm/dd/yyyy hh:mm:ss AM/PM 
     *  depending on the 12/24 format of the iButton. For example:
     *  Wednesday 5/5/1999 3:39:34 PM
     *      
     *  @param registerPage char[] containing a copy of the registerPage from
     *   which the time is determined.
     *  @see #getRegisterPage()
     *  @return a String containing the current time
     */
     public static String getRealTimeClockString(char[] registerPage)
     {
        char[] rtc = new char[7];
        int i;
        for (i=0; i< rtc.length; i++)
        {
            rtc[i] = registerPage[i];
        }
        TimeConverter clockTime = new TimeConverter(rtc);
        return clockTime.toString();
     }
    
    //-------------------------------------------------------------------------
    // getRealTimeClockAlarmString
    /** retrieves and returns the current alarm time from this 1921. The String
     *  is of the form dayofweek hh:mm:ss or dayofweek hh:mm:ss AM/PM for a 
     *  weekly alarm, hh:mm:ss or hh:mm:ss AM/PM for a daily alarm, mm:ss for an 
     *  hourly alarm, and ss for an alarm every minute. The String is empty if 
     *  the alarm is set for every second.
     *  For example:
     *  Wednesday 3:39:34 PM  indicates a weekly alarm 
     *  45:00 indicates an hourly alarm
     *  @see #getRealTimeClockAlarmFrequencyString()
     *  @see #getRealTimeClockAlarmFrequencyMask()
     *  @return a String containing the current alarm time
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public String getRealTimeClockAlarmString()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
         getRegisterPage();
         return getRealTimeClockAlarmString(regPage);
     }
     
    //-------------------------------------------------------------------------
    // getRealTimeClockAlarmString
    /** retrieves and returns the current alarm time from this 1921. The String
     *  is of the form dayofweek hh:mm:ss or dayofweek hh:mm:ss AM/PM for a 
     *  weekly alarm, hh:mm:ss or hh:mm:ss AM/PM for a daily alarm, mm:ss for an 
     *  hourly alarm, and ss for an alarm every minute. The String is empty if 
     *  the alarm is set for every second.
     *  For example:
     *  Wednesday 3:39:34 PM  indicates a weekly alarm 
     *  45:00 indicates an hourly alarm
     *  @see #getRealTimeClockAlarmFrequencyString()
     *  @see #getRealTimeClockAlarmFrequencyMask()
     *  @see #getRegisterPage()
     *  @param registerPage char[] containing a copy of the registerPage from
     *   which the time is determined.
     *  
     *  @return a String containing the current alarm time
     *
     */
     public static String getRealTimeClockAlarmString(char [] registerPage)
     {
        char[] rtca = new char[4];
        int i;
        for (i =0; i< rtca.length; i++)
        {
            rtca[i] = registerPage[i+7];
        }
        
        TimeConverter alarmTime = new TimeConverter(rtca);
        return alarmTime.toString();
    }//end getRealTimeClockAlarmString
    
    //-------------------------------------------------------------------------
    // getRealTimeClockAlarmFrequencyMask
    /** retrieves and returns the current alarm frequency mask from this 1921. 
     *   MS MM MH MD return value
     *   1  1  1  1     0x0F      alarm once per second
     *   0  1  1  1     0x07      alarm when seconds match
     *   0  0  1  1     0x03      alarm when minutes and seconds match
     *   0  0  0  1     0x01      alarm when hours, minutes and seconds match
     *   0  0  0  0     0x00      alarm when day, hours, minutes and seconds match
     *  
     *  @see #getRealTimeClockAlarmFrequencyString()
     *  @see #getRealTimeClockAlarmString()
     *  @return a char containing the mask in the lower 4 bits
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public char getRealTimeClockAlarmFrequencyMask()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
         getRegisterPage();
         return getRealTimeClockAlarmFrequencyMask(regPage);
     }
    
    //-------------------------------------------------------------------------
    // getRealTimeClockAlarmFrequencyMask
    /** retrieves and returns the current alarm frequency mask from this 1921. 
     *   MS MM MH MD return value
     *   1  1  1  1     0x0F      alarm once per second
     *   0  1  1  1     0x07      alarm when seconds match
     *   0  0  1  1     0x03      alarm when minutes and seconds match
     *   0  0  0  1     0x01      alarm when hours, minutes and seconds match
     *   0  0  0  0     0x00      alarm when day, hours, minutes and seconds match
     *  
     *  @see #getRealTimeClockAlarmFrequencyString()
     *  @see #getRealTimeClockAlarmString()
     *  @param registerPage char[] containing a copy of the registerPage from
     *   which the frequency mask is determined.
     *  @see #getRegisterPage()
     *  @return a char containing the mask in the lower 4 bits
     */
     public static char getRealTimeClockAlarmFrequencyMask(char[] registerPage)
     {
         char mask = 0x00;
         char maskBitMask = 0x80;
         mask = (char)(mask + ((registerPage[7] & maskBitMask)>>>7));
         mask = (char)((mask << 1) + ((registerPage[8] & maskBitMask)>>>7));
         mask = (char)((mask << 1) + ((registerPage[9] & maskBitMask)>>>7));
         mask = (char)((mask << 1) + ((registerPage[10] & maskBitMask)>>>7));
         return mask;
     }
     
    //-------------------------------------------------------------------------
    // getRealTimeClockAlarmFrequencyString
    /** retrieves and returns the current alarm frequency mask as a string 
     *  from this 1921. 
     *   MS MM MH MD return value
     *   1  1  1  1     Second    alarm once per second
     *   0  1  1  1     Minute    alarm when seconds match
     *   0  0  1  1     Hour      alarm when minutes and seconds match
     *   0  0  0  1     Day       alarm when hours, minutes and seconds match
     *   0  0  0  0     Week      alarm when day, hours, minutes and seconds match
     *  
     *  @see #getRealTimeClockAlarmFrequencyMask()
     *  @see #getRealTimeClockAlarmString()
     *  @return a String describing how often the alarm will occur
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public String getRealTimeClockAlarmFrequencyString()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
         getRegisterPage();
         return getRealTimeClockAlarmFrequencyString(regPage);
     }
     
    //-------------------------------------------------------------------------
    // getRealTimeClockAlarmFrequencyString
    /** retrieves and returns the current alarm frequency mask as a string 
     *  from this 1921. 
     *   MS MM MH MD return value
     *   1  1  1  1     Second    alarm once per second
     *   0  1  1  1     Minute    alarm when seconds match
     *   0  0  1  1     Hour      alarm when minutes and seconds match
     *   0  0  0  1     Day       alarm when hours, minutes and seconds match
     *   0  0  0  0     Week      alarm when day, hours, minutes and seconds match
     *  
     *  @see #getRealTimeClockAlarmFrequencyMask()
     *  @see #getRealTimeClockAlarmString()
     *  @param registerPage char[] containing a copy of the registerPage from
     *   which the frequency string is determined.
     *  @see #getRegisterPage()
     *  @return a String describing how often the alarm will occur
     *  @exception IllegalArgumentException
     */
     public static String getRealTimeClockAlarmFrequencyString(char[] registerPage)
         throws IllegalArgumentException
     {
         char mask = getRealTimeClockAlarmFrequencyMask(registerPage);
         switch (mask)
         {
            case 0x00:
                return new String("weekly");
            case 0x01:
                return new String("daily");
            case 0x03:
                return new String("hourly");
            case 0x07:
                return new String("once per minute");
            case 0x0F:
                return new String("once per second");
            default:
                throw new IllegalArgumentException("Invalid Alarm Register Mask");
         }//end switch (mask)
     }//end getRealTimeClockAlarmFrequencyString
     
     //-------------------------------------------------------------------------
    // getMissionTimeStamp
    /** retrieves and returns the mission time stamp of this 1921. 
     *  the return string is of the form: mm/dd/yyyy hh:mm
     *  note: if the memory has been cleared, and no new mission 
     *  set up, the time stamp will be: 0//1900 00:00
     *  @return string containing the misssion time stamp
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public String getMissionTimeStamp()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        getRegisterPage();
        return getMissionTimeStamp(regPage);
     }
     
    //-------------------------------------------------------------------------
    // getMissionTimeStamp
    /** retrieves and returns the mission time stamp of this 1921. 
     *  the return string is of the form: mm/dd/yyyy hh:mm
     *  @param registerPage char[] containing a copy of the registerPage from
     *   which the time is determined.
     *  @see #getRegisterPage()
     *  @return string containing the mission time stamp
     */
     public static String getMissionTimeStamp(char[] registerPage)
     {
        char[] mts = new char[5];
        int i;
        for (i =0; i< mts.length; i++)
        {
            mts[i] = registerPage[i+21];
        }
        
        TimeConverter missionTime = new TimeConverter(mts);
        return missionTime.toString();
     }// end getMissionTimeStamp()
    
    
    //--------
    //-------- Temperature Recording Data Access Methods
    //--------
    
    //-------------------------------------------------------------------------
    // getAlarmPagesString
    /** retrieves the alarm time stamp and duration pages  (pages 17 to 19)
     *  @returns String  containing the Low/High designation, time stamp. and 
     *  and duration of each alarm which has occurred. If no alarm a have 
     *  occurred, the String will be empty
     *  The String will be of the format 
     *      # mm/dd/yy hh:mm:ss TO mm/dd/yy hh:mm:ss \n
     *      # mm/dd/yy hh:mm:ss TO mm/dd/yy hh:mm:ss
     *  where # is the alarm desgnation HIGH or LOW, the first day/time is start
     *  and the second is end. Individual alarms are seperated by \n 
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public String getAlarmPagesString()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
    {
        char[] alarmPages = getAlarmPages();
        char[] registerPage = getRegisterPage();
        return getAlarmPagesString(registerPage, alarmPages);
    }
    
    //-------------------------------------------------------------------------
    // getAlarmPagesString
    /** retrieves the alarm time stamp and duration pages  (pages 17 to 19)
     *  @param regPage registerPage from which the start time is obtained
     *  @param alarmPages pages 17-19 of this 1921 which contain the 
     *  recorded temperature alarms
     *  @returns String  containing the Low/High designation, time stamp. and 
     *  and duration of each alarm which has occurred. If no alarm a have 
     *  occurred, the String will be empty
     *  The String will be of the format 
     *      # mm/dd/yy hh:mm:ss TO mm/dd/yy hh:mm:ss \n
     *      # mm/dd/yy hh:mm:ss TO mm/dd/yy hh:mm:ss
     *  where # is the alarm desgnation HIGH or LOW, the first day/time is start
     *  and the second is end. Individual alarms are seperated by \n 
     */
     public static String getAlarmPagesString(char[] regPage, char[] alarmPages)
     {
        boolean lowOccurred, highOccurred;
        int i;
        String retStr = new String();
        
        lowOccurred = getStatusBitTLF(regPage);
        highOccurred = getStatusBitTHF(regPage);
        
        if (!lowOccurred && !highOccurred) //no alarms
            return retStr;
            
        //calculate starting time
        int sampleRate = getSampleRate(regPage);
        int startDelay = getMissionStartDelay(regPage);

        //    get mission time stamp in TimeConverter
        char[] mts = new char[5];
        for(i=0;i<mts.length; i++)
            mts[i] = regPage[i+21];
        TimeConverter alarmTime = new TimeConverter(mts);
        
        alarmTime.addMin(startDelay);
        if (lowOccurred)
        {
            for(i = 0; (i<alarmPages.length) && (i<48); i+=4)
            {
               //for each low alarm
               //calc new time/date
               alarmTime = new TimeConverter(mts);
               
               int numSamples = (int) alarmPages[i+2];
               numSamples = (numSamples << 8) + (int) alarmPages[i+1];
               numSamples = (numSamples << 8) + (int) alarmPages[i];
               if (numSamples == 0)
               {
                  break;
               }
               alarmTime.addMin((numSamples - 1)* sampleRate);
               //add to list
               retStr = retStr + "LOW " + alarmTime + " TO ";
               alarmTime.addMin(((int)alarmPages[i+3]-1) * sampleRate);
               retStr = retStr + " " + alarmTime.toString() + "\n";
            }
        }
        if (highOccurred)
        {
            for(i = 48; (i<alarmPages.length) && (i<96) ; i+=4)
            {
               //for each high alarm
               //calc new time/date
               alarmTime = new TimeConverter(mts);
               
               int numSamples = (int) alarmPages[i+2];
               numSamples = (numSamples << 8) + (int) alarmPages[i+1];
               numSamples = (numSamples << 8) + (int) alarmPages[i];
               if (numSamples == 0)
               {
                   break;
               }
               alarmTime.addMin((numSamples - 1)* sampleRate);
               //add to list
               retStr = retStr + "HIGH " + alarmTime + " TO ";
               alarmTime.addMin(((int)alarmPages[i+3]-1) * sampleRate);
               retStr = retStr + " " + alarmTime + "\n";
            }
        }
       
        return retStr;
     } //end getAlarmPagesString
     
    //-------------------------------------------------------------------------
    // getAlarmPages
    /** retrieves the alarm time stamp and duration pages (pages 17-19) and 
     *  returns them as a character array.
     *  @returns a char[] representation of the alarm pages. Only the lower  
     *   8 bits are used
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public char[] getAlarmPages()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        getRegisterPage();
        char [] alarmPages;
        char [] startingAddr = new char[2];
        boolean lowOccurred, highOccurred;
        int i;
        String retStr = new String();
        
        lowOccurred = getStatusBitTLF(regPage);
        highOccurred = getStatusBitTHF(regPage);
        
        if (!lowOccurred && !highOccurred) //no alarms
            return new char[0];
            
        //read pages
        startingAddr[0] = 0x02;
        startingAddr[1] = 0x20;
        alarmPages = readMemoryWithCRC(startingAddr, 3);
        return alarmPages;
        
     }
    
    //-------------------------------------------------------------------------
    // getHistogramString
    /** retrieves the histogram pages (pages 64 to 67)
     *  @returns String containing the histogram information
     *  The String will be of the format 
     *      startingTemp TO endingTemp, histogramValue
     *  Where the startingTemp and endingTemp are the boundaries of the 
     *  histogram bin, and the histogramValue is the data in the bin.
     *  Individual bins are seperated by \n 
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public String getHistogramString()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        char[] histPages = getHistogramPages();
        return getHistogramString(histPages);
     }
     
    //-------------------------------------------------------------------------
    // getHistogramString
    /** retrieves the histogram pages (pages 64 to 67)
     *  @param histPages pages 64-67 from which the string is created
     *  @returns String containing the histogram information
     *  The String will be of the format 
     *      startingTemp TO endingTemp, histogramValue
     *  Where the startingTemp and endingTemp are the boundaries of the 
     *  histogram bin, and the histogramValue is the data in the bin.
     *  Individual bins are seperated by \n 
     */
     public static String getHistogramString(char[] histPages)
     {
        int i;
        String retStr = new String();
        
        int temp = -40;
        for(i=0; i<histPages.length; i+=2)
        {
            
            int hist = (int)histPages[i+1];
            hist = hist << 8;
            hist = hist + (int)histPages[i];
            retStr = retStr + temp + "\u00b0C TO " + (temp + 1.5) + "\u00b0C, " + hist + "\n";
            temp +=2;
        }
        
        return retStr;
     }//end getHistogramString
     
     //-------------------------------------------------------------------------
    // getHistogramPages
    /** retrieves the histogram pages (pages 64 to 67) of this 1921 and returns 
     *  them in a character array
     *  @returns char[] containing the histogram pages. Only the lower 8 bits of
     *   each character is used
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public char[] getHistogramPages()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException
     {
        char [] histPages;
        char [] startingAddr = new char[2];
        int i;
        String retStr = new String();
        
        //read pages
        startingAddr[0] = 0x08;
        startingAddr[1] = 0x00;
        histPages = readMemoryWithCRC(startingAddr, 4); //pages 64-67
        
        return histPages;
     }
    
    //-------------------------------------------------------------------------
    // getTempLogString
    /** retrieves the temperature logging pages (pages 128 to 191) in a String 
     *  format showing temperature and date/time recorded
     *  NOTE: this method may read up to 64 pages of memory, and will take at 
     *  least 2 seconds to complete
     *  @returns String containing the temperature log information
     *  The String will be of the format 
     *      date/time temp
     *  Where the date/time are of the standard format and the temp is in celsius
     *  Individual readings are seperated by \n 
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public String getTempLogString()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException   
     { 
        char[] logPages = getTempLogPages();
        char[] registerPage = getRegisterPage();
        return getTempLogString(registerPage, logPages);
     }
     
    //-------------------------------------------------------------------------
    // getTempLogString
    /** retrieves the temperature logging pages (pages 128 to 191) in a String 
     *  format showing temperature and date/time recorded
     *  @param regPage register page on which the start time is found
     *  @param logPages pages 128-191 in which the logs are stored to calculate 
     *  the temperature
     *  @returns String containing the temperature log information
     *  The String will be of the format 
     *      date/time temp
     *  Where the date/time are of the standard format and the temp is in celsius
     *  Individual readings are seperated by \n 
     *
      */
     public static String getTempLogString(char[] regPage, char[] tempLog)
     {
        int missionSamples = getMissionSamplesCounter(regPage);
        int sampleRate = getSampleRate(regPage);
        int startDelay = getMissionStartDelay(regPage);
        boolean RO = getControlBitRO(regPage);
        int i, j;
        
        float temp;
        char[] mts = new char[5];
        for(i=0;i<mts.length; i++)
            mts[i] = regPage[i+21];
        TimeConverter logTime = new TimeConverter(mts);
        String retStr = new String();
        
        if ( !RO || (missionSamples <= 2048))
        {
            for (i = 0; (i < missionSamples); i++)
            {
                temp = ((float) tempLog[i]/2)-40;
                retStr = retStr + logTime.toString() + " " + temp + "\n";
                logTime.addMin(sampleRate);
            }
        }
        else //RollOver Has Occurred
        {
            int numRollOvers = missionSamples/2048;
            int earliestLog = missionSamples % 2048;
            logTime.addMin((missionSamples-2048) * sampleRate);
            j = earliestLog;
            for (i = 0; i < 2048; i++)
            {
                temp = ((float) tempLog[j]/2)-40;
                retStr = retStr + logTime.toString() + " " + temp + "\n";
                logTime.addMin(sampleRate);
                j++;
                if (j==2048)
                   j = 0;
            }
        }
        return retStr;
     }//end getTempLogString
     
    //-------------------------------------------------------------------------
    // getTempLogPages
    /** retrieves the temperature logging pages (pages 128 to 191) of this 1921
     *  and  returns them as a character array. If less than 2048 temperature
     *  readings have been taken, the return array will contain only the memory 
     *  pages which have been used. NOTE: this method may read up to 64 pages of 
     *  memory, which would take at least 2 seconds to complete
     *  @returns char[] containing the temperature log information. Only the 
     *  lower 8 bits of each character are used
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */
     public char[] getTempLogPages()
         throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException   
     {
        int missionSamples = getMissionSamplesCounter();
        int i, j;
        
        char[] addr = new char[2];
        addr[0] = 0x10;
        addr[1] = 0x00;
        
        int pagesToRead;
        if (missionSamples > 2048)
            pagesToRead = 64;
        else
        {
            pagesToRead = missionSamples/32;
            if (missionSamples%32 != 0) //if there is a non-full page, read it too
                pagesToRead++;
        }
        
        char[] tempLog = readMemoryWithCRC(addr, pagesToRead);
        
        return tempLog;
     }//get temp log pages

    
    //--------
    //-------- Missioning Methods
    //--------
    
    //-------------------------------------------------------------------------
    // mission
    /** Begins a new mission in the DS1921. This method WILL END any mission in 
     *  progress, clear the memory and the alarms, and set up and start a new
     *  mission with the given parameters. All data from the old mission
     *  will be unretreivable. NOTE: this method takes at least 2 seconds to 
     *  complete.
     *  @param missioningData char[] containing bytes 0x0200 through 0x0214, if 
     *  a new real time clock value is to be sent, or 0x0207 through 0x0214 if 
     *  no new clock value is sent
     *  @param newClock  boolean true if a new Real Time Clock Value is to be 
     *  sent in
     *  0x0214 of the register page which contains all missioning data 
     *  to write to the 1921.
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception IllegalArgumentException
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
     */

    public void mission(char [] missionPage, boolean newClock)
        throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException,
               IllegalArgumentException
    {
        int i;
        int statusReg;
        int controlReg;
        char[] clockAndAlarmData;
        char[] controlDelayStatusData;
        
        
        if (newClock)
        {
            if (missionPage.length != 21)
                throw new IllegalArgumentException("The missioning register pages was " +
                                          missionPage.length + " char long. It "+
                                          "should be 21 char long to mission with a " +
                                          "new Real Time Clock.");
            statusReg = 20;
            controlReg = 14;
        }
        else
        {
            if (missionPage.length != 14)
                throw new IllegalArgumentException("The missioning register pages was " +
                                          missionPage.length + " char long. It "+
                                          "should be 14 char long to mission without a " +
                                          "new Real Time Clock.");
            statusReg = 13;
            controlReg = 7;
        }
        
        getRegisterPage();
        if ( getStatusBitMIP() ) // end current mission
            endMission();
        
        clearMemory(); //clear the memory
       
        missionPage[statusReg] = (char) (missionPage[statusReg] & 0xF0); //clear alarm registers
        missionPage[controlReg] = (char) (missionPage[controlReg] & 0xEF); //enable mission
        
        if (newClock)
        {
            clockAndAlarmData = new char[14];             //data for the real time
            for ( i = 0; i<clockAndAlarmData.length; i++) //clock, clock alarm,  
                clockAndAlarmData[i] = missionPage[i];    //alarm thresholds, and 
                                                          //sample rate
        }
        else
        {
            clockAndAlarmData = new char[7];              //data for the 
            for ( i = 0; i<clockAndAlarmData.length; i++) //clock alarm,  
                clockAndAlarmData[i] = missionPage[i];    //alarm thresholds, and 
                                                          //sample rate
        }
        
        controlDelayStatusData = new char[7];                       //data for the control
        for (i=0; i< controlDelayStatusData.length; i++)            //and status registers
            controlDelayStatusData[i] = missionPage[i+controlReg];  //and the mission start delay
        
        char[] readData;
        char[] addr = new char[2];
        
        //write bytes 0x020E through 0x0214 in the register page
        addr[0] = 0x02;
        addr[1] = 0x0E;
        writeScratchpad(addr, controlDelayStatusData);
        readData = readScratchpad();
        copyScratchpad(readData);
        
        //write bytes 0x0200 through 0x020D or 0x0207 through 0x020D
        //byte 0x020D is the sample rate and when written 
        //will begin the mission countdown
        if (newClock)
        {
            addr[0] = 0x02;
            addr[1] = 0x00;
        }
        else
        {
            addr[0] = 0x02;
            addr[1] = 0x07;
        }
        writeScratchpad(addr, clockAndAlarmData);
        readData = readScratchpad();
        copyScratchpad(readData);
    }


    //-------------------------------------------------------------------------
    //endMission
    /** Ends a currently active mission by writing the Mission In Progress 
     *  Status bit to a 0. All mission data is preserved for reading.
     *
     *  @exception AdapterNotDetectedExcception
     *  @exception MicroLanIOException
     *  @exception PortInUseException
     *  @exception PortNotSelectedException
      */
    public void endMission()
        throws MicroLanIOException, 
               PortInUseException, 
               AdapterNotDetectedException, 
               PortNotSelectedException   
    {
        getRegisterPage();
        char[] sendInfo = new char[1];
        char newMIP = (char) (regPage[20] & 0xDF);
        sendInfo[0] = newMIP;
        char [] addr = new char[2];
        addr[0] = 0x02;
        addr[1] = 0x14;
        writeScratchpad(addr,sendInfo);
        char [] readInfo = readScratchpad();
        if (readInfo[3] != newMIP)
            throw new MicroLanIOException("New MIP not confirmed");
        copyScratchpad(readInfo);
        try { Thread.sleep(1000); } catch (Exception e){} //wait for status to change
        if (getStatusBitMIP()) //mission still in progress
            throw new MicroLanIOException("New MIP not written. Failed to end mission");
    }      

   //--------
   //-------- Temperature format conversion methods
   //--------
   
   //-------------------------------------------------------------------------
   // farenheitToCelsius
   /** converts a floating point farenheit temperature to a floating point 
    *  celsius temperature
    *  @param farenTemp temperature in degrees farenheit
    *  @return float temperature in degrees celsius
    */
    public static float fahrenheitToCelsius(float farenTemp)
    {
        return (float) (( 5 * (farenTemp - 32)) / 9);      
    }
    
   //-------------------------------------------------------------------------
   // DS1921ToCelsius
   /** converts a char representation of a 1921 temperature to a floating point 
    *  celsius temperature
    *  @param DS1921Temp temperature in degrees farenheit
    *  @return float temperature in degrees celsius
    */
    public static float DS1921ToCelsius(char DS1921Temp)
    {
        return (float)DS1921Temp/2 - 40; //convert to celsius    
    }
    
   //-------------------------------------------------------------------------
   // celsiusToDS1921
   /** converts a floating point celsius temperature to a char representation 
    *  of a 1921 temperature.
    *  @param celsiusTemp float temperature in degrees celsius
    *  @return char temperature in DS1921 frormat
    */
    public static char celsiusToDS1921(float celsiusTemp)
    {   
        return (char)(2*celsiusTemp + 80);      
    }
    
    

}//end iButtonContainer21

    //--------
    //-------- Class Time Converter
    //--------
    
    
    //-------------------------------------------------------------------------
    // class TimeConverter
    /** The time converter class provides methods for the conversion of the DS1921
     *  time structure into string format and back. It also provides methods
     *  which allow increments of time to be added to an initial time in order to 
     *  provide time stamps for alarms and temperature readings. This class assumes 
     *  that Sunday = 1, and that the century bit = 0 means 1900, as reccommended 
     *  in the DS1921 data sheet.
     */
    class TimeConverter
    {
        int sec;
        int min;
        int hr;
        int day;
        int date;
        int mo;
        int yr;
        int type;
        boolean cent;
        boolean twelveHour;
               
        /** Constant time format real time clock */
        public final int RTC = 0;
        /** Constant time format real time clock alarm */
        public final int RTCA = 1;
        /** Constant time format mission start time */
        public final int MST = 2;
        
        /** length of each month in days */
        private final int[] monthLength = {0,31,29,31,30,31,30,31,31,30,31,30,31};;
        
        /** array of month name abbreviations */
        String[] abbrMonth = {new String(), "Jan.", "Feb.", "Mar.", "Apr.", "May" , "Jun.",
                              "Jul.", "Aug.", "Sept.", "Oct.", "Nov.", "Dec."};
       /** array of day name abbreviations */
        String[] abbrDay ={new String(), "Sun.", "Mon.", "Tues.", "Wed.", "Thurs.", "Fri.", "Sat."};
        
        //---------------------------------------------------------------------------
        //--------Constructors
        //---------------------------------------------------------------------------
        
        /** creates a TimeConverter object from a DS1921 format time char[]*/
        public TimeConverter(char[] time)
        {
            switch (time.length)
             {
                case 4:
                    convertRTCA(time);
                    break;
                case 5: 
                    convertMST(time);
                    break;
                case 7: 
                    convertRTC(time);
                    break;
                default:
                    sec = 0;
                    min = 0;
                    hr = 0;
                    day = 0;
                    date = 0;
                    mo = 0;
                    yr = 0;
                    type = 0;
                    cent = false;
                    twelveHour= false;
                }
        }//end constructor
        
        
        /** creates a time converter object from a Mission Time Stamp char[]*/
        private void convertMST(char[] time)
        {
            type = MST;
            setMinutes(time[0]);
            setHours(time[1]);
            day = 0;
            setDate(time[2]);
            setMonth(time[3]);
            
            char[] years = new char[2];
            years[0] = time[3];
            years[1] = time[4];
            
            setYear(years);
        }
        
        /** creates a TimeConverter Object from a DS1921 Real Time Clock Alarm char[]*/
        private void convertRTCA(char[] time)
        {
            type = RTCA;
            setSeconds(time[0]);
            setMinutes(time[1]);
            setHours(time[2]);
            setDay(time[3]);
            
        }
        
        /** creates a TimeConverter Object from a DS1921 Real Time Clock char[]*/
        private void convertRTC(char[] time)
        {
            type = RTC;
            setSeconds(time[0]);
            setMinutes(time[1]);
            setHours(time[2]);
            setDay(time[3]);
            setDate(time[4]);
            setMonth(time[5]);
            
            char[] years = new char[2];
            years[0] = time[5];
            years[1] = time[6];
            
            setYear(years);
        }
        
        /**Sets the seconds from the second byte*/
        private void setSeconds(char sec)
        {
            this.sec = ( sec & 0x0F ) + (((sec >>> 4) & 0x07)*10);
        }
        
        /**Gets a ds1921 format second char from a given number of seconds*/
        private char getSeconds(int sec)
            throws IllegalArgumentException
        {
            if ((sec <0) || (sec>59))
                throw new IllegalArgumentException("Seconds out of range. Range is 0-59");
            int ones = sec % 10;
            int tens = sec / 10;
            
            char seconds = (char) ( (tens << 4) | ones);
            return seconds;
        }
        
        /**Sets the minutes from the minutes byte*/
        private void setMinutes(char min)
        {
            this.min = ( min & 0x0F ) + (((min >>> 4) & 0x07)*10);
        }
        
        /**Gets a ds1921 format minutes char from an integer*/
        private char getMinutes(int min)
            throws IllegalArgumentException
        {
            if ((min <0) || (min>59))
                throw new IllegalArgumentException("Minutes out of range. Range is 0-59");
            int ones = min % 10;
            int tens = min / 10;
            
            char minutes = (char) ( (tens << 4) | ones);
            return minutes;
        }
        
        /**Sets the hours and 12 hour flag from the hours byte*/ 
        private void setHours(char hours)
        {
             twelveHour = ((hours & 0x40) == 0x40);
              
             hr = (hours & 0x0F);
             
             if (twelveHour)
             {
                hr = hr + (((hours >>> 4) & 0x01)*10);
                if ((hours & 0x20) == 0x20) //PM
                    hr = hr+12;
             }
             else //24 hour
             {
                hr = hr + (((hours >>> 4) & 0x03)*10);
             }
        }
        
        /**Gets a ds1921 hours byte from an hour int, 12 hour boolean, and PM boolean*/
        private char getHours(int hours, boolean hr12, boolean PM)
            throws IllegalArgumentException
        {
            if ((hours > 24) || (hr12 && (hours >12)))
                throw new IllegalArgumentException("hours out of range. Range is 0-23 " +
                                                    "for 24 hour time, and 0-12 for 12 " +
                                                    "hour time.");
            
            int tens = hours / 10;
            int ones = hours % 10;
            int hourByte = 0 | ((tens << 4) | ones) ;
            if (hr12)
            {
                hourByte = hourByte | 0x40;
                if (PM)
                {
                    hourByte = hourByte | 0x20;
                }
            }
            else //24 hour time
            {}
            return (char)hourByte;
        }
               
        
        /** sets the day of the week from the day of week byte*/
        private void setDay(char day)
        {
            this.day = (int) day & 0x07;
        }
        
        /** gets the ds1921 byte for day of the week from a String */
        private char getDay(String dayString)
            throws IllegalArgumentException
        {
            dayString = dayString.toUpperCase();
            char dayChar;
            if (dayString.startsWith("SU"))
                dayChar = 0x01;
            else if (dayString.startsWith("MO"))
                dayChar = 0x02;
            else if (dayString.startsWith("TU"))
                dayChar = 0x03;
            else if (dayString.startsWith("WE"))
                dayChar = 0x04;
            else if (dayString.startsWith("TH"))
                dayChar = 0x05;
            else if (dayString.startsWith("FR"))
                dayChar = 0x06;
            else if (dayString.startsWith("SA"))
                dayChar = 0x07;
            else
                throw new IllegalArgumentException("String does not indicate a day of the week.");
                
            return dayChar;
        }//end getDay
        
        /** sets the date from the date byte */
        private void setDate(char date)
        {
            this.date = ((date >>> 4) *10) | (date & 0x0F);
        }
            
        /** gets a DS1921 format date char from the date, month, and  year ints */
        private char getDate(int date, int month, int year)
            throws IllegalArgumentException
        {
            if ((date <0 ) || (month < 0) || (year <0))
                throw new IllegalArgumentException("Date, month and year must all be positive");
            if ( date > monthLength[month])
                throw new IllegalArgumentException("Date " +date+ " is not in month " + month);
            if ((month ==2) && (date == 29))
                if ( !((year%400)==0) && !(((year%100)==0) && ((year%4)==0)))
                     throw new IllegalArgumentException("Date " +date+ " is not in month " + month + " of " + year);
           
            //date OK
            int ones = date % 10;
            int tens = date / 10;
            
            char dateChar = (char) ((tens << 4) | ones);
            
            return dateChar;
        }//end get month
            
            
        /** sets the months from the months byte*/
        private void setMonth( char month)
        {
            mo = (month & 0x0F) + (((month >>> 4) &0x01) *10);
        }
        
        
        /** gets a DS1921 month from months int, and cent boolean */
        private char getMonth(int month, boolean cent)
        {
            if ((month <1) || (month >12))
                throw new IllegalArgumentException("Month out of range. range is 1-12");
            
            int ones = month % 10;
            int tens = month /10;
            
            char monthChar = (char) ((tens << 4) | ones);
            
            if (cent)
                monthChar = (char)(monthChar | 0x80);
            
            return monthChar;
        }//end get month
        
        //** sets the year from the months and year bytes */
        private void setYear(char[] year)
            throws IllegalArgumentException
        {
            if (year.length != 2)
                throw new IllegalArgumentException("Incorrect length for year array. " +
                                    "Should consist of exactly two bytes. ");
            
            cent = (year[0] & 0x80) == 0x80;
             
            yr = + (year[1]&0x0F) + (((year[1] & 0xF0) >>> 4)* 10);
             
            if (cent)
                yr = yr + 2000;
            else 
                yr = yr + 1900;
        }
        
        //** gets a DS1921 format year from the year int */
        private char getYear(int year)
            throws IllegalArgumentException
        {
            if (year < 0)
                throw new IllegalArgumentException("Year must be positive");
            int ones = year % 10;
            int tens = (year % 100) /10;
            
            char yearChar = (char) ((tens << 4) | ones);
            
            return yearChar;
        }
        
        /** converts the time information into a string */
        public String toString()
        {
            String retStr = new String();
            if((type == RTC) || (type == RTCA))
            {
               retStr = retStr + abbrDay[day];
               if (!abbrDay[day].equals(new String()))
                   retStr=  retStr + " ";
            }
            if((type == RTC) || (type == MST))
               retStr = retStr + abbrMonth[mo] + " ";
            if((type == RTC) || (type == MST))
               retStr = retStr + date + ",";
            if((type == RTC) || (type == MST))
               retStr = retStr + yr + " ";
               
            if((twelveHour) && (hr >12))
                retStr = retStr + (hr-12) + ":";
            else
                retStr =  retStr + hr + ":";
                
            if(min <=9)
                retStr = retStr+ "0";
            retStr = retStr + min;
            
            if((type == RTC) || (type == RTCA))
            {
                retStr = retStr + ":";
                if (sec<=9)
                    retStr = retStr+"0";
                retStr = retStr + sec;
            }
            
            if ((twelveHour) && (hr >12))
                retStr = retStr + " PM";
            else if (twelveHour)
                retStr = retStr + " AM";
            
            return retStr;
        }//end toString
        
        /** adds the specified number of minutes to the time */    
        public void addMin(int x)
        {
            min = min + x%60;
            x = x/60;
            hr = hr + min/60 + x%24;
            min = min%60;
            x = x/24;
            day = day + (hr/24 + x)%7;
            if (day > 7)
                day = day%7;
            date = date + hr/24 + x;
            hr = hr%24;
            
            if (type == RTCA)
            {
                date = 0;
                return;
            }
            
            
            boolean fixMonth = true;
            while(fixMonth)
            {
               if ( mo == 2)
                {
                    fixMonth = !FebOK();
                    System.out.println("     Interim date at: " + toString());
                }
                else if (date > monthLength[mo])
                {
                    date = date - monthLength[mo];
                    mo++;
                    if (mo >= 13)
                    {
                        mo = mo - 12;
                        yr++;
                        if (yr%2 == 0)
                           cent = true;
                        else 
                           cent = false;
                    }
                    if ((date > monthLength[mo]) || (mo == 2))
                        fixMonth = true;
                    else 
                        fixMonth = false;
                }//end else if date >  monthLength[mo];
                else
                    fixMonth = false;
            }//end while fixMonth
        }//end add minutes
         
        /** Determines whether a given date in February is valid or not */
        public boolean FebOK()
        {
           if (date <=28)
               return true;
           else //date is >28
           {
               if ( (yr % 400 == 0) || ( (yr%100 != 0) && (yr%4 == 0)))
               {
                   //leapyear
                   if (date <= 29)
                       return true;
                   else
                   {
                       date = date - 29;
                       mo++;
                       return false;
                   }
               }
               else //not a leap year, and date is > 28
               {
                   date = date -28;
                   mo++;
                   return false;
               }
           }
        }
        
        /** creates a Real Time Clock array from the time string*/
        public char[] toRTCTime(String time)
            throws IllegalArgumentException 
        {
            char[] dsTime = new char[7];
            StringTokenizer st = new StringTokenizer(time,"\n\t\r/: ");
            int i = 0;
            String dayOfWeek = null;
            int seconds = 0;
            int minutes = 0;
            int hours = 0;
            int date = 0;
            int month = 0;
            int year = 0;
            boolean PM = false;
            boolean twelve = false;
            
            while (st.hasMoreTokens() && i<8)
            {
                switch (i)
                {
                    case 0:
                       dayOfWeek = st.nextToken();
                       break;
                    case 1:
                       month = Integer.parseInt(st.nextToken());
                       break;
                    case 2:
                       date = Integer.parseInt(st.nextToken());
                       break;
                    case 3:
                       year = Integer.parseInt(st.nextToken());
                       break;
                    case 4: 
                       hours = Integer.parseInt(st.nextToken());
                       break;
                    case 5:
                       minutes = Integer.parseInt(st.nextToken());
                       break;
                    case 6:
                       seconds = Integer.parseInt(st.nextToken());
                       break;
                    case 7: //there is an AM/PM symbol
                       twelve = true;
                       PM = st.nextToken().equals("PM");
                       break;
                    default:
                }
                i++;
            }
            if (i<7)
                 throw new IllegalArgumentException("Incorrect number of time/date parameters");
            try
            {
                dsTime[0] = getSeconds(seconds);
                dsTime[1] = getMinutes(minutes);
                dsTime[2] = getHours(hours,twelve,PM);
                dsTime[3] = getDay(dayOfWeek);
                dsTime[4] = getDate(date,month,year);
                boolean century;
                if (year >= 2000)
                    century = true;
                else
                    century = false;
                dsTime[5] = getMonth(month,century);
                dsTime[6] = getYear(year);
            }
            catch (Exception e)
            {
                 e.printStackTrace();
            }
            return dsTime;
        }//end to RTC time
        
        
        /** creates a Real Time Clock Alarm array from the time string*/
        public char[] toRTCATime(String time, char alarmMask)
            throws IllegalArgumentException 
        {
            char[] dsTime = new char[4];
            StringTokenizer st = new StringTokenizer(time,"\n\t\r/: ");
            int i = 0;
            String dayOfWeek = null;
            int seconds = 0;
            int minutes = 0;
            int hours = 0;
            boolean twelve = false;
            boolean PM = false;
            
            while (st.hasMoreTokens() && i<5)
            {
                switch (i)
                {
                    case 0:
                       dayOfWeek = st.nextToken();
                       break;
                    case 1:
                       hours = Integer.parseInt(st.nextToken());
                       break;
                    case 2:
                       minutes = Integer.parseInt(st.nextToken());
                       break;
                    case 3:
                       seconds = Integer.parseInt(st.nextToken());
                       break;
                    case 4: // AM/PM symbol present
                       twelve = true;
                       PM = st.nextToken().equals("PM");
                    default:
                }
                i++;
            }
            if (i<4)
                 throw new IllegalArgumentException("Incorrect number of time/date parameters");
            try
            {
                dsTime[0] = getSeconds(seconds);
                if ((alarmMask & 0x08) != 0)
                    dsTime[0] = (char) ( dsTime[0] | 0x80);
                dsTime[1] = getMinutes(minutes);
                if ((alarmMask & 0x04) != 0)
                    dsTime[1] = (char) ( dsTime[1] | 0x80);
                dsTime[2] = getHours(hours,twelve,PM);
                if ((alarmMask & 0x02) != 0)
                    dsTime[2] = (char) ( dsTime[2] | 0x80);
                dsTime[3] = getDay(dayOfWeek);
                if ((alarmMask & 0x01) != 0)
                    dsTime[3] = (char) ( dsTime[3] | 0x80);
            }
            catch (Exception e)
            {
                 e.printStackTrace();
            }
            return dsTime;
        }//end to RTCA time
        
        /** creates a Mission Start Time array from the time string*/
        public char[] toMSTTime(String time)
            throws IllegalArgumentException 
        {
            char[] dsTime = new char[5];
            StringTokenizer st = new StringTokenizer(time,"\n\t\r/: ");
            int i = 0;
            int minutes = 0;
            int hours = 0;
            int date = 0;
            int month = 0;
            int year = 0;
            boolean twelve = false;
            boolean PM = false;
            
            while (st.hasMoreTokens() && i<6)
            {
                switch (i)
                {
                    case 0:
                       month = Integer.parseInt(st.nextToken());
                       break;
                    case 1:
                       date = Integer.parseInt(st.nextToken());
                       break;
                    case 2:
                       year = Integer.parseInt(st.nextToken());
                       break;
                    case 3: 
                       hours = Integer.parseInt(st.nextToken());
                       break;
                    case 4:
                       minutes = Integer.parseInt(st.nextToken());
                       break;
                    case 5: //there is an AM/PM symbol
                       twelve = true;
                       PM = st.nextToken().equals("PM");
                       break;
                    default:
                }
                i++;
            }
            if (i<5)
                 throw new IllegalArgumentException("Incorrect number of time/date parameters");
            try
            {
                dsTime[0] = getMinutes(minutes);
                dsTime[1] = getHours(hours,twelve,PM);
                dsTime[2] = getDate(date,month,year);
                boolean century;
                if (year >= 2000)
                    century = true;
                else
                    century = false;
                dsTime[3] = getMonth(month,century);
                dsTime[4] = getYear(year);
            }
            catch (Exception e)
            {
                 e.printStackTrace();
            }
            return dsTime;
        }//end to MST time
        
    }//end class timeConverter
